#' Use rolling windows quantiles to calculate local quantiles of R|W

#' @param r Values of R
#' @param w Values of W
#' @param tau value in (0,1) - find the tau-quantile
#' @param J number of windows for rolling window estimate
#' @param wqg Gap between the quantiles of w to calculate the sizes of the rolling windows
#'
#' @return List with elements "w.midpoint",
#'  containing the midpoints of the w values of each rolling window,
#'  "r.tau.w" giving an estimate of the tau-quantile for R|W for that window,
#'  and "r0w" containing a threshold value of R for each w.
#'
#' @author Jenny Wadsworth



empiricalQR.2d<-function(r,w,tau=0.95,J=60,wqg=0.08,quantilespacing=TRUE)
{
  wsl<-seq(0,1-wqg,len=J)
  wsu<-seq(wqg,1,len=J)
  lowerwquants<-quantile(w,wsl)
  upperwquants<-quantile(w,wsu)

  lowerwquants[1]<-0.99*lowerwquants[1]
  upperwquants[J]<-1.01*upperwquants[J]

  if(!quantilespacing){
    lowerwquants<-wsl
    upperwquants<-wsu
  }

  qu.r<-NULL
  for(j in 1:J)
  {
    re<-r[lowerwquants[j]<w & w<upperwquants[j]]
    qu.r[j]<-quantile(re,tau)
  }

  ##################
  n<-length(w)
  r0w<-numeric(n)
  for(i in 1:n)
  {
    indvec<-rep(NA,J)
    for(j in 1:J)
    {
      indvec[j]<-lowerwquants[j]<w[i] & w[i]<upperwquants[j]
    }
    r0w[i]<-mean(qu.r[indvec],na.rm = T)
  }
  ##################

  return(list(w.midpoint=as.numeric((lowerwquants+upperwquants)/2), r.tau.w=qu.r, r0w=r0w))
}


#===============================================================

#' Use rolling windows to calculate local quantiles of R|W

#' @param r Values of R
#' @param w1,w2 Values of W1,W2
#' @param tau value in (0,1) - find the tau-quantile
#' @param J number of windows for rolling window estimate
#' @param wg Gap between the values of w to calculate the sizes of the rolling windows
#'
#' @return List with elements "w.midpoint", a J^2 x 2 matrix containing the midpoints of the w1,w2 values of each rolling window, "r.tau.w", a J x J matrix giving an estimate of the tau-quantile for R|(W_1,W_2) for each window, and "r0w", a threshold value of R|(W1,W2) for each value of (w1,w2)
#'
#' @author Jenny Wadsworth



empiricalQR.3d<-function(r,w1,w2,tau=0.95,J=40,wg=0.1)
{
  wsl<-seq(0,1-wg,len=J)
  wsu<-seq(wg,1,len=J)

  qu.r<-matrix(0,J,J)

  for(j in 1:J)
  {
    for(k in 1:J)
    {
      ind<-wsl[j]<w1 & w1<wsu[j] & wsl[k]<w2 & w2<wsu[k]
      r1<-r[ind]
      if(sum(ind)>10)
      {
        qu.r[j,k]<-quantile(r1,tau)
      }
      else{qu.r[j,k]<-NA}
    }
  }

  ########
  # Find r0w for each value of w (for use in likelihood functions)
  n<-length(w1)
  r0w<-numeric(n)
  for(i in 1:n)
  {
    indMat<-matrix(NA,J,J)
    for(j in 1:J)
    {
      for(k in 1:J)
      {
        indMat[j,k]<-wsl[j]<w1[i] & w1[i]<wsu[j] & wsl[k]<w2[i] & w2[i]<wsu[k]
      }
    }
    r0w[i]<-mean(qu.r[indMat],na.rm = T)
  }
  ########

  return(list(w.midpoint=expand.grid((wsl+wsu)/2,(wsl+wsu)/2), r.tau.w=qu.r, r0w=r0w))
}


#====================================================================

#' Plot empirical estimate of gauge function based on empirical quantile regression

# @param w.midpoint Mid points of the rolling W windows used to calculate quantiles of R|W - output from empiricalQR.2d
# @param r.tau.w Estimates of the (high) tau-quantiles of R|W - output from empiricalQR.2d
#' @param wpts w values used to calculate quantiles of R|W - output from QR.2d
#' @param r.tau.wpts Estimates of the (high) tau-quantiles of R|W corresponding to W=wpts - output from QR.2d
#' @param points logical - if TRUE then plot points
#' @param lines logical - if TRUE then lines join the points
#' @param return.coords logical - if TRUE then returns a matrix of x-y coordinates used to generate the plot
#' @param add logical - if TRUE, then add points / lines to an existing plot
#' @return Plot of an empirical estimate of the gauge function
#'
#' @author Jenny Wadsworth
#' @export


# plotempiricalgauge.2d<-function(w.midpoint,r.tau.w, lines=TRUE, return.coords=FALSE)
# {
#   k1<-max(w.midpoint*r.tau.w, na.rm=TRUE)
#   k2<-max((1-w.midpoint)*r.tau.w, na.rm=TRUE)
#   plot(w.midpoint*r.tau.w/k1, (1-w.midpoint)*r.tau.w/k2, pch=20, xlab="x",ylab="y",xlim=c(0,1),ylim=c(0,1))
#   if(lines)
#   {
#     lines(w.midpoint*r.tau.w/k1, (1-w.midpoint)*r.tau.w/k2)
#   }
#   if(return.coords)
#   {
#     return(cbind(w.midpoint*r.tau.w/k1, (1-w.midpoint)*r.tau.w/k2))
#   }
# }

plotempiricalgauge.2d<-function(wpts,r.tau.wpts, points=TRUE, lines=TRUE, return.coords=FALSE, add=FALSE)
{
  k1<-max(wpts*r.tau.wpts, na.rm=TRUE)
  k2<-max((1-wpts)*r.tau.wpts, na.rm=TRUE)

  if(add)
  {
    if(points)
    {
    points(wpts*r.tau.wpts/k1, (1-wpts)*r.tau.wpts/k2, pch=20)
    }
    if(lines)
    {
      lines(wpts*r.tau.wpts/k1, (1-wpts)*r.tau.wpts/k2)
    }
  }
  else{
    if(points)
    {
    plot(wpts*r.tau.wpts/k1, (1-wpts)*r.tau.wpts/k2, pch=20, xlab="x",ylab="y",xlim=c(0,1),ylim=c(0,1))
      if(lines)
      {
        lines(wpts*r.tau.wpts/k1, (1-wpts)*r.tau.wpts/k2)
      }
    } else if(!points && lines)
    {
      plot(wpts*r.tau.wpts/k1, (1-wpts)*r.tau.wpts/k2, typ="l", xlab="x",ylab="y",xlim=c(0,1),ylim=c(0,1))
    }
  }
  if(return.coords)
  {
    return(cbind(wpts*r.tau.wpts/k1, (1-wpts)*r.tau.wpts/k2))
  }
}


#' Plot empirical estimate of gauge function based on empirical quantile regression

#' @param w.midpoint Mid points of the rolling (W1,W2) windows used to calculate quantiles of R|(W1,W2) - output from empiricalQR.3d
#' @param r.tau.w Estimates of the (high) tau-quantiles of R|(W1,W2) - output from empiricalQR.3d
#' @param surface logical - if TRUE then adds surface rendering to the points in the 3d plot
#' @param return.coords logical - if TRUE then returns a matrix of x-y-z coordinates used to generate the plot
#' @return 3d Plot of an empirical estimate of the gauge function
#'
#' @author Jenny Wadsworth
#' @export

plotempiricalgauge.3d<-function(w.midpoint,r.tau.w, surface=TRUE, return.coords=FALSE)
{
  k1<-max(w.midpoint[,1]*r.tau.w, na.rm=TRUE)
  k2<-max(w.midpoint[,2]*r.tau.w, na.rm=TRUE)
  k3<-max((1-w.midpoint[,1]-w.midpoint[,2])*r.tau.w, na.rm=TRUE)

  rgl::plot3d(w.midpoint[,1]*r.tau.w/k1, w.midpoint[,2]*r.tau.w/k2, pmax((1-w.midpoint[,1]-w.midpoint[,2])*r.tau.w/k3,0), pch=20, xlab="x",ylab="y",zlab="z")
  if(surface)
  {
    rgl::surface3d(w.midpoint[,1]*r.tau.w/k1, w.midpoint[,2]*r.tau.w/k2, pmax((1-w.midpoint[,1]-w.midpoint[,2])*r.tau.w/k3,0),col=2,alpha=0.5,add=T)
  }
  if(return.coords)
  {
    return(cbind(c(w.midpoint[,1]*r.tau.w/k1), c(w.midpoint[,2]*r.tau.w/k2), pmax(c((1-w.midpoint[,1]-w.midpoint[,2])*r.tau.w/k3),0)))
  }
}


#====================================================================

#' Use rolling windows quantiles or additive quantile regression to calculate local quantiles of R|W

#' @param r Values of R
#' @param w Values of W
#' @param method Character string: "empirical" for empirical rolling-windows quantiles, "smooth" for generalized additive quantile regression using the R library qgam
#' @param tau value in (0,1) - find the tau-quantile
#' @param J number of windows for rolling window estimate
#' @param wqg Gap between the quantiles / values of w to calculate the sizes of the rolling windows
#' @param quantilespacing logical. If TRUE then size of rolling windows is determined by the quantiles of W. If FALSE, rolling windows are of a fixed length equal to wqg
#' @param k.smooth Number of basis functions to use if method="smooth"
#' @param bs.smooth Type of basis functions to use if method="smooth" (See qgam library for further details).
#'
#'
#' @return List with three elements (i) "wpts",
#'  containing w values: for method="empirical", these are the midpoints of the w values of each rolling window; for method="smooth", these are values on a grid.
#'  (ii) "r.tau.wpts" giving an estimate of the tau-quantile for R|W corresponding to each of the w values in wpts,
#'  and (iii) "r0w" containing a threshold value of R for each w in the data.
#'
#' @author Jenny Wadsworth
#' @export


QR.2d<-function(r,w,method=c("empirical","smooth"),tau=0.95,J=60,wqg=0.08,quantilespacing=TRUE, k.smooth=15,bs.smooth="tp")
{
  if(method=="smooth")
  {
    require(qgam)
    fit <- qgam(r~s(w,  k=k.smooth, bs=bs.smooth), data = as.data.frame(cbind(r,w)), qu = tau)
    r0w <- predict(fit, newdata = as.data.frame(cbind(r,w)))

    excind<-r>r0w
    rexc<-r[excind]
    wexc<-w[excind]
    r0wexc<-r0w[excind]

    wpts<-seq(0, 1, length.out = 1000)
    rwSeq <- data.frame(cbind(r = rep(0, 1000), w = wpts))
    r.tau.wpts <- predict(fit, newdata = rwSeq)
  }
  else{
    eqr<-empiricalQR.2d(r=r,w=w,tau=tau,J=J,wqg=wqg,quantilespacing=quantilespacing)
    wpts<-eqr$w.midpoint
    r.tau.wpts<-eqr$r.tau.w
    r0w<-eqr$r0w
  }
  # Return three objects:
  #   1. Sequence of w points and
  #   2. tau-quantiles of R|W=w for this sequence, for plotting an empirical gauge
  #   3. values of r_0(w) for each value of w supplied in the data

  return(list(wpts=wpts, r.tau.wpts=r.tau.wpts, r0w=r0w))
}



#===============================================================

#' Use rolling windows to calculate local quantiles of R|W

#' @param r Values of R
#' @param w Values of W: should either be a two-column matrix containing (W1,W2) or a three-column matrix containing (W1,W2,W3)
#' @param method Character string: "empirical" for empirical rolling-windows quantiles. This is the only implemented method at present.
#' @param tau value in (0,1) - find the tau-quantile
#' @param J number of windows for rolling window estimate
#' @param wg Gap between the values of w to calculate the sizes of the rolling windows
#'
#' @return List with elements "wpts", a J^2 x 2 matrix containing the midpoints of the w1,w2 values of each rolling window, "r.tau.wpts", a J x J matrix giving an estimate of the tau-quantile for R|(W_1,W_2) for each window, and "r0w", a threshold value of R|(W1,W2) for each value of (w1,w2) in the data
#'
#' @author Jenny Wadsworth
#' @export


QR.3d<-function(r,w,method="empirical",tau=0.95,J=40,wg=0.1)
{
  w1<-w[,1]
  w2<-w[,2]
  eqr<-empiricalQR.3d(r=r,w1=w1,w2=w2,tau=tau,J=J,wg=wg)
  wpts<-eqr$w.midpoint
  r.tau.wpts<-eqr$r.tau.w
  r0w<-eqr$r0w

  return(list(wpts=wpts, r.tau.wpts=r.tau.wpts, r0w=r0w))
}



#===============================================================

#' Calculate Euclidian distance between two vectors

euclid.dist<-function(x1,x2){return(sqrt(sum((x1-x2)^2)))}

#===============================================================

#' Check on vertices of (rectangular) set of interest to see if it lies in region R'>k (R>k*r_0(W))
#'
#' @param x,y each should be a 2-vector defining the vertices of the rectangle
#' @param w all W values for the dataset under consideration
#' @param r0W values of r_0(W) corresponding to all W values for the dataset under consideration
#' @param k value of k to check
#'
#' @return logical value: TRUE if rectangular set lies in region or FALSE if not. NB there is a small degree of approximation involved.
#'
#' @author Jenny Wadsworth


checkset.2d<-function(x,y,w,r0w,k=1)
{
  x11<-x[1]; x12<-x[2]
  y11<-y[1]; y12<-y[2]

  if(is.null(dim(w))){w<-cbind(w,1-w)}

  i1<-which.min(apply(w,1,euclid.dist,x2=c(x11,y11)/(x11+y11)))
  i2<-which.min(apply(w,1,euclid.dist,x2=c(x12,y11)/(x12+y11)))
  i3<-which.min(apply(w,1,euclid.dist,x2=c(x11,y12)/(x11+y12)))
  i4<-which.min(apply(w,1,euclid.dist,x2=c(x12,y12)/(x12+y12)))

  logic<-all((x11+y11)>k*r0w[i1],
             (x12+y11)>k*r0w[i2],
             (x11+y12)>k*r0w[i3],
             (x12+y12)>k*r0w[i4])

  return(logic)
}

#===============================================================

#' Check on vertices of (cuboid) set of interest to see if it lies in region R'>k (R>k*r_0(W))
#'
#' @param x,y,z each should be a 2-vector defining the vertices of the cuboid
#' @param w all W values for the dataset under consideration
#' @param r0W values of r_0(W) corresponding to all W values for the dataset under consideration
#' @param k value of k to check
#'
#' @return logical value: TRUE if cuboid set lies in region or FALSE if not. NB there is a small degree of approximation involved.
#'
#' @author Jenny Wadsworth


checkset.3d<-function(x,y,z,w,r0w,k=1)
{
  x11<-x[1]; x12<-x[2]
  y11<-y[1]; y12<-y[2]
  z11<-z[1]; z12<-z[2]

  if(dim(w)[2]==2){w<-cbind(w,1-w[,1]-w[,2])}

  i1<-which.min(apply(w,1,euclid.dist,x2=c(x11,y11,z11)/(x11+y11+z11)))
  i2<-which.min(apply(w,1,euclid.dist,x2=c(x12,y11,z11)/(x12+y11+z11)))
  i3<-which.min(apply(w,1,euclid.dist,x2=c(x11,y12,z11)/(x11+y12+z11)))
  i4<-which.min(apply(w,1,euclid.dist,x2=c(x11,y11,z12)/(x11+y11+z12)))
  i5<-which.min(apply(w,1,euclid.dist,x2=c(x12,y12,z11)/(x12+y12+z11)))
  i6<-which.min(apply(w,1,euclid.dist,x2=c(x12,y11,z12)/(x12+y11+z12)))
  i7<-which.min(apply(w,1,euclid.dist,x2=c(x11,y12,z12)/(x11+y12+z12)))
  i8<-which.min(apply(w,1,euclid.dist,x2=c(x12,y12,z12)/(x12+y12+z12)))

  logic<-all((x11+y11+z11)>k*r0w[i1],
             (x12+y11+z11)>k*r0w[i2],
             (x11+y12+z11)>k*r0w[i3],
             (x11+y11+z12)>k*r0w[i4],
             (x12+y12+z11)>k*r0w[i5],
             (x12+y11+z12)>k*r0w[i6],
             (x11+y12+z12)>k*r0w[i7],
             (x12+y12+z12)>k*r0w[i8])

  return(logic)
}
