#' Simulate new data (using resampled W|R'>k) from the model, conditional on R'>k (R>k*r_0(W))

#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param k Value of k (defaults to 1). If k>1 then importance weights are calculated to simulate from the distribution of X|R>kr_0(W) (equivalently X|R'>k)
#' @param nsim Number of new points to simulate
#' @param par The vector of fitted parameters (in the same order as obtained from fit.geometricMVE.2d)
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return nsim by 2 matrix of values (in exponential margins) simulated from the model (given R>r_0(W), or R>kr_0(W) for k>1)
#'
#' @author Jenny Wadsworth
#' @export
#'

sim.2d<-function(w,r0w,k=1,nsim,par,gfun,add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{

  if(k!=1){
    if(add.gauge)
    {
      iw<-iweights.2d(k=k, r0w=r0w, w=w, par=par,add.gauge=add.gauge,gauge1=gauge1,gauge2=gauge2,par1ind=par1ind,par2ind=par2ind)
    } else{
      iw<-iweights.2d(k=k, r0w=r0w, w=w, gfun=gfun, par=par,add.gauge=add.gauge)
    }
    star.ind<-sample(1:length(w),size=nsim,replace=T,prob = iw)
    wstar<-w[star.ind]
    r0w_star<-c(k*r0w)[star.ind]
  }else{
    star.ind<-sample(1:length(w),size=nsim,replace=T)
    wstar<-w[star.ind]
    r0w_star<-r0w[star.ind]
  }

  if(!add.gauge){
    rate0<-apply(cbind(w,1-w),1,gfun,par=par[2])
    rate<-rate0[star.ind]

    rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
                                       shape=par[1],rate=rate,
                                       lower.tail=F),
                  shape=par[1],rate=rate)
  }
  else{
    ms<-additivegauge.scaling(gauge1=gauge1,gauge2=gauge2,weight=par[4],par1=par[2],par2=par[3])
    rate0<-apply(cbind(w,1-w),1,additivegauge.rescale,par1=par[2],par2=par[3],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[4])
    rate<-rate0[star.ind]
    rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
                                       shape=par[1],
                                       rate=rate,
                                       lower.tail=F),
                  shape=par[1],rate=rate)
  }

  xstar<-cbind(rstar*wstar,rstar*(1-wstar))
  return(xstar)
}


# sim.2d<-function(w,r0w,k=1,nsim,par,gfun=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square),
#                  add.gauge=FALSE,gauge1=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square),
#                  gauge2=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square))
# {
#
#   if(k!=1){
#     iw<-iweights(k=k, r0w=r0w, w=w, gfun=gfun, par=par)
#     star.ind<-sample(1:length(w),size=nsim,replace=T,prob = iw)
#     wstar<-w[star.ind]
#     r0w_star<-c(k*r0w)[star.ind]
#   }else{
#   star.ind<-sample(1:length(w),size=nsim,replace=T)
#   wstar<-w[star.ind]
#   r0w_star<-r0w[star.ind]
#   }
#
#   if(!add.gauge){
#     rate0<-apply(cbind(w,1-w),1,gfun,par=par[2])
#     rate<-rate0[star.ind]
#
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                        shape=par[1],rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#   else{
#     ms<-additivegauge.scaling(gauge1=gauge1,gauge2=gauge2,weight=par[4],par1=par[2],par2=par[3])
#     rate0<-apply(cbind(w,1-w),1,additivegauge.rescale,par1=par[2],par2=par[3],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[4])
#     rate<-rate0[star.ind]
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                        shape=par[1],
#                                        rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#
#   xstar<-cbind(rstar*wstar,rstar*(1-wstar))
#   return(xstar)
# }

# sim.2d<-function(w,r0w,nsim,par,gfun=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square),
#                  add.gauge=FALSE,gauge1=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square),
#                  gauge2=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square))
# {
#
#   star.ind<-sample(1:length(w),size=nsim,replace=T)
#   wstar<-w[star.ind]
#   r0w_star<-r0w[star.ind]
#
#   if(!add.gauge){
#     # rate<-apply(cbind(wstar,1-wstar),1,gfun,par=par[2])
#     rate0<-apply(cbind(w,1-w),1,gfun,par=par[2])
#     rate<-rate0[star.ind]
#
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                        shape=par[1],rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#   else{
#     ms<-additivegauge.scaling(gauge1=gauge1,gauge2=gauge2,weight=par[4],par1=par[2],par2=par[3])
#     #  rate<-apply(cbind(wstar,1-wstar),1,additivegauge.rescale,par1=par[2],par2=par[3],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[4])
#     rate0<-apply(cbind(w,1-w),1,additivegauge.rescale,par1=par[2],par2=par[3],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[4])
#     rate<-rate0[star.ind]
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                        shape=par[1],
#                                        rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#
#   xstar<-cbind(rstar*wstar,rstar*(1-wstar))
#   return(xstar)
# }

# sim.2d<-function(w,r0w,nsim,par,gfun=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square),
#                  add.gauge=FALSE,gauge1=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square),
#                  gauge2=c(gauge_rvad, gauge_gaussian,gauge_invlogistic, gauge_square))
# {
#
#    star.ind<-sample(1:length(w),size=nsim,replace=T)
#    wstar<-w[star.ind]
#    r0w_star<-r0w[star.ind]
#
#    if(!add.gauge){
#    rate<-apply(cbind(wstar,1-wstar),1,gfun,par=par[2])
#    rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                   shape=par[1],rate=rate,
#                                   lower.tail=F),
#                  shape=par[1],rate=rate)
#    }
#    else{
#     ms<-additivegauge.scaling(gauge1=gauge1,gauge2=gauge2,weight=par[4],par1=par[2],par2=par[3])
#     rate<-apply(cbind(wstar,1-wstar),1,additivegauge.rescale,par1=par[2],par2=par[3],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[4])
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                       shape=par[1],
#                                       rate=rate,
#                                       lower.tail=F),
#                    shape=par[1],rate=rate)
#    }
#
#    xstar<-cbind(rstar*wstar,rstar*(1-wstar))
#    return(xstar)
# }



#=========================================================

#' Simulate new data (using resampled W|R'>k) from the model, conditional on R'>k (R>k*r_0(W))

#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param k Value of k (defaults to 1). If k>1 then importance weights are calculated to simulate from the distribution of X|R>kr_0(W) (equivalently X|R'>k)
#' @param nsim Number of new points to simulate
#' @param par The vector of fitted parameters (in the same order as obtained from fit.geometricMVE.2d)
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return nsim by 3 matrix of values (in exponential margins) simulated from the model (given R>r_0(W), or R>kr_0(W) for k>1)
#'
#' @author Jenny Wadsworth
#' @export
#'

sim.3d<-function(w,r0w,k=1,nsim,par,gfun,
                 add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{
  if(k!=1){
    if(add.gauge)
    {
      iw<-iweights.3d(k=k, r0w=r0w, w=w, par=par,add.gauge=add.gauge,gauge1=gauge1,gauge2=gauge2,par1ind=par1ind,par2ind=par2ind)
    } else{
      iw<-iweights.3d(k=k, r0w=r0w, w=w, gfun=gfun, par=par,add.gauge=add.gauge)
    }
    star.ind<-sample(1:length(w[,1]),size=nsim,replace=T,prob = iw)
    wstar<-w[star.ind,]
    r0w_star<-c(k*r0w)[star.ind]
  }else{
    if(dim(w)[2]==2){w<-cbind(w,1-w[,1]-w[,2])}
    star.ind<-sample(1:length(w[,1]),size=nsim,replace=T)
    wstar<-w[star.ind,]
    r0w_star<-r0w[star.ind]
  }

  if(!add.gauge){
    # rate<-apply(wstar,1,gfun,par=par[2:length(par)])
    rate0<-apply(w,1,gfun,par=par[2:length(par)])
    rate<-rate0[star.ind]
    rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
                                       shape=par[1],rate=rate,
                                       lower.tail=F),
                  shape=par[1],rate=rate)
  }
  else{
    wgrid<-create.wgrid.3d(70)
    ms<-additivegauge.scaling.3d(gauge1=gauge1,gauge2=gauge2,weight=par[length(par)],par1=par[par1ind],par2=par[par2ind],wgrid=wgrid)
    rate0<-apply(w,1,additivegauge.rescale.3d,
                 par1=par[par1ind],par2=par[par2ind],ms=ms,
                 gauge1=gauge1,gauge2=gauge2,weight=par[length(par)])
    rate<-rate0[star.ind]
    rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,shape=par[1],
                                       rate=rate,
                                       lower.tail=F),
                  shape=par[1],rate=rate)
  }

  xstar<-rstar*wstar
  return(xstar)
}



# sim.3d<-function(w,r0w,k=1,nsim,par,gfun,
#                  add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
# {
#   if(k!=1){
#     iw<-iweights(k=k, r0w=r0w, w=w, gfun=gfun, par=par)
#     star.ind<-sample(1:length(w[,1]),size=nsim,replace=T,prob = iw)
#     wstar<-w[star.ind,]
#     r0w_star<-c(k*r0w)[star.ind]
#   }else{
#   if(dim(w)[2]==2){w<-cbind(w,1-w[,1]-w[,2])}
#   star.ind<-sample(1:length(w[,1]),size=nsim,replace=T)
#   wstar<-w[star.ind,]
#   r0w_star<-r0w[star.ind]
#   }
#
#   if(!add.gauge){
#     # rate<-apply(wstar,1,gfun,par=par[2:length(par)])
#     rate0<-apply(w,1,gfun,par=par[2:length(par)])
#     rate<-rate0[star.ind]
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                        shape=par[1],rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#   else{
#     wgrid<-create.wgrid.3d(70)
#     ms<-additivegauge.scaling.3d(gauge1=gauge1,gauge2=gauge2,weight=par[length(par)],par1=par[par1ind],par2=par[par2ind],wgrid=wgrid)
#     rate0<-apply(w,1,additivegauge.rescale.3d,
#                  par1=par[par1ind],par2=par[par2ind],ms=ms,
#                  gauge1=gauge1,gauge2=gauge2,weight=par[length(par)])
#     rate<-rate0[star.ind]
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,shape=par[1],
#                                        rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#
#   xstar<-rstar*wstar
#   return(xstar)
# }

# sim.3d<-function(w,r0w,nsim,par,gfun,
#                  add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
# {
#   if(dim(w)[2]==2){w<-cbind(w,1-w[,1]-w[,2])}
#   star.ind<-sample(1:length(w[,1]),size=nsim,replace=T)
#   wstar<-w[star.ind,]
#   r0w_star<-r0w[star.ind]
#
#   if(!add.gauge){
#     rate<-apply(wstar,1,gfun,par=par[2:length(par)])
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,
#                                        shape=par[1],rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#   else{
#     wgrid<-create.wgrid.3d(70)
#     ms<-additivegauge.scaling.3d(gauge1=gauge1,gauge2=gauge2,weight=par[length(par)],par1=par[par1ind],par2=par[par2ind],wgrid=wgrid)
#     rate<-apply(w,1,additivegauge.rescale.3d,
#                 par1=par[par1ind],par2=par[par2ind],ms=ms,
#                 gauge1=gauge1,gauge2=gauge2,weight=par[length(par)])
#     rstar<-qgamma(1-runif(nsim)*pgamma(r0w_star,shape=par[1],
#                                        rate=rate,
#                                        lower.tail=F),
#                   shape=par[1],rate=rate)
#   }
#
#   xstar<-rstar*wstar
#   return(xstar)
# }

#==============================================================================
#' Calculate importance weights for simulation of W|R'>k, k>1, (equivalently W|R>k r_0(W), where R'=R/r_0(W)).

#' @param k value of k>1: the importance weights are to simulate from the distribution of W|R'>k
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun fitted gauge function
#' @param par fitted parameter values (first value should be the gamma shape parameter, subsequent values are the gauge function parameters)
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return vector of (unnormalized) importance weights to be used in sim.2d
#'
#' @author Jenny Wadsworth

iweights.2d<-function(k, r0w, w, gfun, par, add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{
  if(k<1){stop("k should exceed 1")}
  if(is.null(dim(w)[2])){w<-cbind(w,1-w)} # For d=2
  if(add.gauge){
    ms<-additivegauge.scaling(gauge1=gauge1,gauge2=gauge2,weight=par[length(par)],par1=par[par1ind],par2=par[par2ind])
    rate<-apply(w,1,additivegauge.rescale,par1=par[par1ind],par2=par[par2ind],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[length(par)])
  }
  else{
    rate<-apply(w,1,gfun,par=par[2:length(par)])
  }
  return(pgamma(k*r0w,shape=par[1],rate=rate,lower.tail = F)/pgamma(r0w,shape=par[1],rate=rate,lower.tail = F))
}

#==============================================================================
#' Calculate importance weights for simulation of W|R'>k, k>1, (equivalently W|R>k r_0(W), where R'=R/r_0(W)).

#' @param k value of k>1: the importance weights are to simulate from the distribution of W|R'>k
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun fitted gauge function
#' @param par fitted parameter values (first value should be the gamma shape parameter, subsequent values are the gauge function parameters)
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return vector of (unnormalized) importance weights to be used in sim.3d
#'
#' @author Jenny Wadsworth

iweights.3d<-function(k, r0w, w, gfun, par, add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{
  if(k<1){stop("k should exceed 1")}
  if(dim(w)[2]==2){w<-cbind(w,1-w[,1]-w[,2])} # For d=3 if w is only 2 column
  if(add.gauge){
    wgrid<-create.wgrid.3d(70)
    ms<-additivegauge.scaling.3d(gauge1=gauge1,gauge2=gauge2,weight=par[length(par)],par1=par[par1ind],par2=par[par2ind],wgrid=wgrid)
    rate<-apply(w,1,additivegauge.rescale.3d,par1=par[par1ind],par2=par[par2ind],ms=ms,gauge1=gauge1,gauge2=gauge2,weight=par[length(par)])
  }
  else{
    rate<-apply(w,1,gfun,par=par[2:length(par)])
  }
  return(pgamma(k*r0w,shape=par[1],rate=rate,lower.tail = F)/pgamma(r0w,shape=par[1],rate=rate,lower.tail = F))
}

# iweights<-function(k, r0w, w, gfun, par)
# {
#   if(k<1){stop("k should exceed 1")}
#   if(is.null(dim(w)[2])){w<-cbind(w,1-w)} # For d=2
#   else if(dim(w)[2]==2){w<-cbind(w,1-w[,1]-w[,2])} # For d=3 if w is only 2 column
#   rate<-apply(w,1,gfun,par=par[2:length(par)])
#   return(pgamma(k*r0w,shape=par[1],rate=rate,lower.tail = F)/pgamma(r0w,shape=par[1],rate=rate,lower.tail = F))
# }

#==============================================================================
#' Calculate P(R'>k|R'>1) (equivalently P(R>kr_0(W)|R>r_0(w)))

#' @param k value of k>1
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun fitted gauge function
#' @param par fitted parameter values (first value should be the gamma shape parameter, subsequent values are the gauge function parameters)
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return Estimate of P(R'>k|R'>1)
#'
#' @author Jenny Wadsworth
#' @export

Rexc.prob.k.2d<-function(k, r0w, w, gfun, par, add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{
  iw<-iweights.2d(k, r0w, w, gfun, par, add.gauge,gauge1,gauge2,par1ind,par2ind)
  return(mean(iw))
}

#==============================================================================

#' Calculate P(R'>k|R'>1) (equivalently P(R>kr_0(W)|R>r_0(w)))

#' @param k value of k>1
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun fitted gauge function
#' @param par fitted parameter values (first value should be the gamma shape parameter, subsequent values are the gauge function parameters)
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return Estimate of P(R'>k|R'>1)
#'
#' @author Jenny Wadsworth
#' @export
#'
Rexc.prob.k.3d<-function(k, r0w, w, gfun, par, add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{
  iw<-iweights.3d(k, r0w, w, gfun, par, add.gauge,gauge1,gauge2,par1ind,par2ind)
  return(mean(iw))
}


#
#
# Rexc.prob.k<-function(k, r0w, w, gfun, par)
# {
#   iw<-iweights(k=k, r0w=r0w, w=w, gfun=gfun, par=par)
#   return(mean(iw))
# }

