# PP and QQ plots

#' PP diagnostic for fitted truncated gamma model

#' @param r Values of R|W that exceed the threshold r0w
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param par The vector of fitted parameters (in the same order as obtained from fit.geometricMVE.2d)
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#'
#' @return PP plot with 95\% tolerance intervals
#'
#' @author Jenny Wadsworth and Ryan Campbell
#' @export

ppdiag.4d = function (r, w, r0w, par, gfun, add.gauge = FALSE, gauge1, gauge2,
                      par1ind, par2ind)
{
  if (dim(w)[2] == 3) {
    w <- cbind(w, 1 - w[, 1] - w[, 2] - w[, 3])
  }
  if (!add.gauge) {
    num <- pgamma(r, shape = par[1], rate = apply(w, 1,
                                                  gfun, par = par[2:length(par)]), lower.tail = F)
    den <- pgamma(r0w, shape = par[1], rate = apply(w, 1,
                                                    gfun, par = par[2:length(par)]), lower.tail = F)
  }
  else {
    wgrid <- create.wgrid.4d(20)
    ms <- additivegauge.scaling.4d(gauge1 = gauge1, gauge2 = gauge2,
                                   weight = par[length(par)], par1 = par[par1ind],
                                   par2 = par[par2ind], wgrid = wgrid)
    num <- pgamma(r, shape = par[1], rate = apply(w, 1,
                                                  additivegauge.rescale.4d, par1 = par[par1ind], par2 = par[par2ind],
                                                  ms = ms, gauge1 = gauge1, gauge2 = gauge2, weight = par[length(par)]),
                  lower.tail = F)
    den <- pgamma(r0w, shape = par[1], rate = apply(w, 1,
                                                    additivegauge.rescale.4d, par1 = par[par1ind], par2 = par[par2ind],
                                                    ms = ms, gauge1 = gauge1, gauge2 = gauge2, weight = par[length(par)]),
                  lower.tail = F)
  }
  nr <- length(r)
  plot(c(1:nr)/(nr + 1), sort(1 - num/den), xlab = "empirical",
       ylab = "model", pch = 20, cex = 0.8)
  abline(a = 0, b = 1)
  Ulow <- sapply(1:nr, function(i) {
    qbeta(0.025, i, nr + 1 - i)
  })
  Uup <- sapply(1:nr, function(i) {
    qbeta(0.975, i, nr + 1 - i)
  })
  lines(c(1:nr)/(nr + 1), Ulow, lty = 2)
  lines(c(1:nr)/(nr + 1), Uup, lty = 2)
}

#' QQ diagnostic for fitted truncated gamma model - puts PP plot on any desired quantile scale (default is standard exponential)

#' @param r Values of R|W that exceed the threshold r0w
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param par The vector of fitted parameters (in the same order as obtained from fit.geometricMVE.2d)
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param quantilefn a quantile function to specify the desired scale. Defaults to qexp.
#'
#' @return QQ plot with 95\% tolerance intervals
#'
#' @author Jenny Wadsworth and Ryan Campbell
#' @export

qqdiag.4d = function (r, w, r0w, par, gfun, add.gauge = FALSE, gauge1, gauge2,
                      par1ind, par2ind, quantilefn = qexp)
{
  if (dim(w)[2] == 3) {
    w <- cbind(w, 1 - w[, 1] - w[, 2] - w[, 3])
  }
  if (!add.gauge) {
    num <- pgamma(r, shape = par[1], rate = apply(w, 1,
                                                  gfun, par = par[2:length(par)]), lower.tail = F)
    den <- pgamma(r0w, shape = par[1], rate = apply(w, 1,
                                                    gfun, par = par[2:length(par)]), lower.tail = F)
  }
  else {
    wgrid <- create.wgrid.4d(20)
    ms <- additivegauge.scaling.4d(gauge1 = gauge1, gauge2 = gauge2,
                                   weight = par[length(par)], par1 = par[par1ind],
                                   par2 = par[par2ind], wgrid = wgrid)
    num <- pgamma(r, shape = par[1], rate = apply(w, 1,
                                                  additivegauge.rescale.4d, par1 = par[par1ind], par2 = par[par2ind],
                                                  ms = ms, gauge1 = gauge1, gauge2 = gauge2, weight = par[length(par)]),
                  lower.tail = F)
    den <- pgamma(r0w, shape = par[1], rate = apply(w, 1,
                                                    additivegauge.rescale.4d, par1 = par[par1ind], par2 = par[par2ind],
                                                    ms = ms, gauge1 = gauge1, gauge2 = gauge2, weight = par[length(par)]),
                  lower.tail = F)
  }
  nr <- length(r)
  plot(quantilefn(c(1:nr)/(nr + 1)), quantilefn(sort(1 - num/den)),
       xlab = "empirical", ylab = "model", pch = 20, cex = 0.8)
  abline(a = 0, b = 1)
  Ulow <- sapply(1:nr, function(i) {
    qbeta(0.025, i, nr + 1 - i)
  })
  Uup <- sapply(1:nr, function(i) {
    qbeta(0.975, i, nr + 1 - i)
  })
  lines(quantilefn(c(1:nr)/(nr + 1)), quantilefn(Ulow), lty = 2)
  lines(quantilefn(c(1:nr)/(nr + 1)), quantilefn(Uup), lty = 2)
}


#====================================================================

# Empirical Quantile Regression

#' Use rolling windows quantiles to calculate local quantiles of R|W

#' @param r Values of R
#' @param w Values of W
#' @param tau value in (0,1) - find the tau-quantile
#' @param J number of windows for rolling window estimate
#' @param wqg Gap between the quantiles of w to calculate the sizes of the rolling windows
#'
#' @return List with elements "w.midpoint",
#'  containing the midpoints of the w values of each rolling window,
#'  "r.tau.w" giving an estimate of the tau-quantile for R|W for that window,
#'  and "r0w" containing a threshold value of R for each w.
#'
#' @author Jenny Wadsworth and Ryan Campbell

empiricalQR.4d = function(r, w, tau = 0.95, J = 20, wg = 0.1){

  w1 = w[,1]
  w2 = w[,2]
  w3 = w[,3]

  # part 1: get 0.95 (tau) quantile of r in each gridpoint
  wsl <- seq(0, 1 - wg, len = J)
  wsu <- seq(wg, 1, len = J)
  qu.r <- array(0, dim=c(J,J,J))
  for (j in 1:J) {
    for (k in 1:J) {
      for (l in 1:J){
        ind <- wsl[j] < w1 & w1 < wsu[j] & wsl[k] < w2 & w2 < wsu[k] & wsl[l] < w3 & w3 < wsu[l]
        r1 <- r[ind]
        if (sum(ind) > 10) {
          qu.r[j, k, l] <- quantile(r1, tau)
        }
        else {
          qu.r[j, k, l] <- NA
        }
      }
    }
  }

  # part 2, compute r0w for each set of w's
  n <- length(w1)
  r0w <- numeric(n)
  for (i in 1:n) {
    indArray <- array(NA, dim=c(J,J,J))
    for (j in 1:J) {
      for (k in 1:J) {
        for (l in 1:J) {
          indArray[j, k, l] <- wsl[j] < w1[i] & w1[i] < wsu[j] & wsl[k] < w2[i] & w2[i] < wsu[k] & wsl[l] < w3[i] & w3[i] < wsu[l]
        }
      }
    }
    r0w[i] <- mean(qu.r[indArray], na.rm = T)
  }

  return(list(w.midpoint = expand.grid((wsl + wsu)/2, (wsl + wsu)/2, (wsl + wsu)/2),
              r.tau.w = qu.r,
              r0w = r0w))
}

#====================================================================

# Gauge functions

gauge_gaussian4d = function (vec, par)
{

  S <- matrix(c(1, par[1:3],
                par[1], 1, par[4:5],
                par[2], par[4], 1, par[6],
                par[3], par[5], par[6], 1),
              4, 4, byrow = T)
  return(t(sqrt(vec)) %*% solve(S) %*% sqrt(vec))
}

# gauge.rvad.4d.river.flow
# =================

# For fitting a model on data obtained from sites A, B, C, and D on the river flow dataset
# Where asymptotic dependence was determined to occur on the faces 2, 4, (1,4), (1,3,4), and (1,2,3,4)

# gauge.rvad.4d.river.flow = function(input.vec, par, th14=1, th134=1, th1234=1){
#
#   if(length(par)!=sum(th14[th14==1],th134[th134==1],th1234[th1234==1])){
#     warning("Incorrect number of parameters")
#     return(NULL)
#   }
#
#   k = 1
#   if(th14==1){
#     al14 = par[k]
#     k = k+1
#   } else {
#     al14 = 1
#   }
#   if(th134==1){
#     al134 = par[k]
#     k = k+1
#   } else {
#     al134 = 1
#   }
#   if(th1234==1){
#     al1234 = par[k]
#     k = k+1
#   } else {
#     al1234 = 1
#   }
#
#   x1 = input.vec[1]
#   x2 = input.vec[2]
#   x3 = input.vec[3]
#   x4 = input.vec[4]
#
#   m14 = min(x1,x4)
#   m134 = min(x1,x3,x4)
#   m1234 = min(x1,x2,x3,x4)
#
#   V1 = c(((x1/al134) + (1 - (1/al134))*m134)[th134==1],
#          ((x1/al1234) + (1 - (1/al1234))*m1234)[th1234==1])
#
#   V2 = c(x2,
#          ((x2/al14) + (1 - (1/al14))*m14)[th14==1],
#          ((x2/al134) + (1 - (1/al134))*m134)[th134==1],
#          ((x2/al1234) + (1 - (1/al1234))*m1234)[th1234==1])
#
#   V3 = c(((x3/al14) + (1 - (1/al14))*m14)[th14==1],
#          ((x3/al134) + (1 - (1/al134))*m134)[th134==1],
#          ((x3/al1234) + (1 - (1/al1234))*m1234)[th1234==1])
#
#   V4 = c(((x4/al1234) + (1 - (1/al1234))*m1234)[th1234==1])
#
#   V12 = c((((x1+x2)/al134) + (1 - (2/al134))*m134)[th134==1],
#           (((x1+x2)/al1234) + (1 - (2/al1234))*m1234)[th1234==1])
#
#   V13 = c((((x1+x3)/al134) + (1 - (2/al134))*m134)[th134==1],
#           (((x1+x3)/al1234) + (1 - (2/al1234))*m1234)[th1234==1])
#
#   V14 = c((((x1+x4)/al1234) + (1 - (2/al1234))*m1234)[th1234==1])
#
#   V23 = c((((x2+x3)/al14) + (1 - (2/al14))*m14)[th14==1],
#           (((x2+x3)/al134) + (1 - (2/al134))*m134)[th134==1],
#           (((x2+x3)/al1234) + (1 - (2/al1234))*m1234)[th1234==1])
#
#   V24 = c((((x2+x4)/al1234) + (1 - (2/al1234))*m1234)[th1234=1])
#
#   V34 = c((((x3+x4)/al1234) + (1 - (2/al1234))*m1234)[th1234==1])
#
#   V123 = c((((x1+x2+x3)/al134) + (1 - (3/al134))*m134)[th134==1],
#            (((x1+x2+x3)/al1234) + (1 - (3/al1234))*m1234)[th1234==1])
#
#   V124 = c((((x1+x2+x4)/al1234) + (1 - (3/al1234))*m1234)[th1234==1])
#
#   V134 = c((((x1+x3+x4)/al1234) + (1 - (3/al1234))*m1234)[th1234==1])
#
#   V234 = c((((x2+x3+x4)/al1234) + (1 - (3/al1234))*m1234)[th1234==1])
#
#   V1234 = c((((x1+x2+x3+x4)/al1234) + (1 - (4/al1234))*m1234)[th1234==1])
#
#   term1 = V1234
#   suppressWarnings(term2 <- min(apply(expand.grid(V123, V4, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term3 <- min(apply(expand.grid(V124, V3, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term4 <- min(apply(expand.grid(V12, V34, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term5 <- min(apply(expand.grid(V12, V3, V4, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term6 <- min(apply(expand.grid(V134, V2, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term7 <- min(apply(expand.grid(V13, V24, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term8 <- min(apply(expand.grid(V13, V2, V4, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term9 <- min(apply(expand.grid(V14, V23, KEEP.OUT.ATTRS = F),
#                                       1, function(x) { sum(x) })))
#   suppressWarnings(term10 <- min(apply(expand.grid(V1, V234, KEEP.OUT.ATTRS = F),
#                                        1, function(x) { sum(x) })))
#   suppressWarnings(term11 <- min(apply(expand.grid(V1, V23, V4, KEEP.OUT.ATTRS = F),
#                                        1, function(x) { sum(x) })))
#   suppressWarnings(term12 <- min(apply(expand.grid(V14, V2, V3, KEEP.OUT.ATTRS = F),
#                                        1, function(x) { sum(x) })))
#   suppressWarnings(term13 <- min(apply(expand.grid(V1, V24, V3, KEEP.OUT.ATTRS = F),
#                                        1, function(x) { sum(x) })))
#   suppressWarnings(term14 <- min(apply(expand.grid(V1, V2, V34, KEEP.OUT.ATTRS = F),
#                                        1, function(x) { sum(x) })))
#
#   return(min(term1, term2, term3, term4, term5, term6, term7, term8,
#              term9, term10, term11, term12, term13, term14))
# }

#====================================================================

#' Negative log likelihood function for truncated gamma model for R|W (4-dimensional data)

#' @param psi vector of parameters. First component is the gamma shape parameter
#' @param r Values of R|W=w that exceed the threshold r0w
#' @param w Values of W that correspond to the threshold exceedances of R (3 or 4 column matrix)
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in psi
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in psi
#' @param pos.par logical. Set to TRUE if all parameters are positive (avoids some warnings in optimization).
#' @param lower.limit optional vector of lower limits on parameters (should be same length as psi)
#' @param upper.limit optional vector of upper limits on parameters (should be same length as psi)
#' @param wgrid 4 column matrix containing points that cover the 3-d simplex. Used only if add.gauge=T, to perform numerical rescaling

#' @return negative log-likelihood value
#'
#' @author Jenny Wadsworth and Ryan Campbell

rw.gamma.lik.4d = function (psi, r, w, r0w, gfun, add.gauge = FALSE, gauge1, gauge2,
                           par1ind, par2ind, pos.par, lower.limit, upper.limit, wgrid)
{
  if (pos.par) {
    if (any(psi <= 0)) {
      return(1e+11)
    }
  }
  if(!is.null(lower.limit)){if(any(psi<lower.limit)){return(10e10)}}
  if(!is.null(upper.limit)){if(any(psi>upper.limit)){return(10e10)}}

  if (dim(w)[2] == 3) {
    w <- cbind(w[, 1], w[, 2], w[, 3], 1 - w[, 1] - w[, 2] - w[, 3])
  }
  if (add.gauge) {
    ms <- additivegauge.scaling.4d(gauge1 = gauge1, gauge2 = gauge2,
                                   weight = psi[length(psi)], par1 = psi[par1ind],
                                   par2 = psi[par2ind], wgrid = wgrid)
    rate <- apply(w, 1, additivegauge.rescale.4d, par1 = psi[par1ind],
                  par2 = psi[par2ind], ms = ms, gauge1 = gauge1, gauge2 = gauge2,
                  weight = psi[length(psi)])
    ll1 <- dgamma(r, shape = psi[1], rate = rate, log = T)
    ll2 <- log(pgamma(r0w, shape = psi[1], rate = rate,
                      lower.tail = F))
  }
  else {
    rate <- apply(w, 1, gfun, par = psi[2:length(psi)])
    ll1 <- dgamma(r, shape = psi[1], rate = rate, log = T)
    ll2 <- log(pgamma(r0w, shape = psi[1], rate = rate, lower.tail = F))
  }
  return(-sum(ll1) + sum(ll2))
}

#' Fit truncated gamma model for R|W for 4-dimensional data

#' @param r Values of R|W=w that exceed the threshold r0w
#' @param w Values of W that correspond to the threshold exceedances of R (3 or 4 column matrix)
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in psi
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in psi
#' @param init.val vector of initial parameter values for optimization
#' @param pos.par logical. Set to TRUE if all parameters are positive (avoids some warnings in optimization).
#' @param lower.limit optional vector of lower limits on parameters (should be same length as init.val)
#' @param upper.limit optional vector of upper limits on parameters (should be same length as init.val)
#' @param ... Additional arguments to pass to optim
#'
#' @return list containing elements "mle", "nllh", "convergence", and "cov" if hessian is set to TRUE
#'
#' @author Jenny Wadsworth and Ryan Campbell
#' @export

fit.geometricMVE.4d = function (r, w, r0w, gfun, add.gauge = FALSE, gauge1, gauge2,
                                par1ind = NULL, par2ind = NULL, init.val = NULL, pos.par = TRUE,
                                lower.limit=NULL, upper.limit=NULL,...)
{
  if (add.gauge && (is.null(par1ind) || is.null(par2ind))) {
    stop("Please specify positions of parameters for gauge1 and gauge2")
  }
  if (add.gauge) {
    wgrid <- create.wgrid.4d(20)
  }
  opt <- optim(rw.gamma.lik.4d, par = init.val, r = r, w = w,
               r0w = r0w, gfun = gfun, add.gauge = add.gauge, gauge1 = gauge1,
               gauge2 = gauge2, pos.par = pos.par, lower.limit=lower.limit, upper.limit=upper.limit,
               par1ind = par1ind, par2ind = par2ind, wgrid = wgrid, ...)
  if (is.null(opt$hessian)) {
    return(list(mle = opt$par, nllh = opt$value, convergence = opt$conv))
  }
  else {
    return(list(mle = opt$par, cov = solve(opt$hessian),
                nllh = opt$value, convergence = opt$conv))
  }
}

#============================================================================

#' Simulate new data (using resampled W|R'>k) from the model, conditional on R'>k (R>k*r_0(W))

#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param k Value of k (defaults to 1). If k>1 then importance weights are calculated to simulate from the distribution of X|R>kr_0(W) (equivalently X|R'>k)
#' @param nsim Number of new points to simulate
#' @param par The vector of fitted parameters (in the same order as obtained from fit.geometricMVE.2d)
#' @param gfun gauge function to be used in the likelihood. Can use inbuilt examples or specify own function of the form gauge<-function(xy,par){}. Leave blank if want to mix gauges additively.
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return nsim by 4 matrix of values (in exponential margins) simulated from the model (given R>r_0(W), or R>kr_0(W) for k>1)
#'
#' @author Jenny Wadsworth and Ryan Campbell
#' @export
#'

sim.4d = function (w, r0w, k = 1, nsim, par, gfun, add.gauge = FALSE,
                   gauge1, gauge2, par1ind, par2ind){
  if (k != 1) {
    if (add.gauge) {
      iw <- iweights.4d(k = k, r0w = r0w, w = w, par = par,
                        add.gauge = add.gauge, gauge1 = gauge1, gauge2 = gauge2,
                        par1ind = par1ind, par2ind = par2ind)
    }
    else {
      iw <- iweights.4d(k = k, r0w = r0w, w = w, gfun = gfun,
                        par = par, add.gauge = add.gauge)
    }
    star.ind <- sample(1:length(w[, 1]), size = nsim, replace = T,
                       prob = iw)
    wstar <- w[star.ind, ]
    r0w_star <- c(k * r0w)[star.ind]
  }
  else {
    if (dim(w)[2] == 3) {
      w <- cbind(w, 1 - w[, 1] - w[, 2] - w[, 3])
    }
    star.ind <- sample(1:length(w[, 1]), size = nsim, replace = T)
    wstar <- w[star.ind, ]
    r0w_star <- r0w[star.ind]
  }
  if (!add.gauge) {
    rate0 <- apply(w, 1, gfun, par = par[2:length(par)])
    rate <- rate0[star.ind]
    rstar <- qgamma(1 - runif(nsim) * pgamma(r0w_star, shape = par[1],
                                             rate = rate, lower.tail = F), shape = par[1], rate = rate)
  }
  else {
    wgrid <- create.wgrid.4d(20)
    ms <- additivegauge.scaling.4d(gauge1 = gauge1, gauge2 = gauge2,
                                   weight = par[length(par)], par1 = par[par1ind],
                                   par2 = par[par2ind], wgrid = wgrid)
    rate0 <- apply(w, 1, additivegauge.rescale.4d, par1 = par[par1ind],
                   par2 = par[par2ind], ms = ms, gauge1 = gauge1, gauge2 = gauge2,
                   weight = par[length(par)])
    rate <- rate0[star.ind]
    rstar <- qgamma(1 - runif(nsim) * pgamma(r0w_star, shape = par[1],
                                             rate = rate, lower.tail = F), shape = par[1], rate = rate)
  }
  xstar <- rstar * wstar
  return(xstar)
}

#' Calculate importance weights for simulation of W|R'>k, k>1, (equivalently W|R>k r_0(W), where R'=R/r_0(W)).

#' @param k value of k>1: the importance weights are to simulate from the distribution of W|R'>k
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun fitted gauge function
#' @param par fitted parameter values (first value should be the gamma shape parameter, subsequent values are the gauge function parameters)
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return vector of (unnormalized) importance weights to be used in sim.4d
#'
#' @author Jenny Wadsworth and Ryan Campbell

iweights.4d = function (k, r0w, w, gfun, par, add.gauge = FALSE, gauge1, gauge2,
                        par1ind, par2ind)
{
  if (k < 1) {
    stop("k should exceed 1")
  }
  if (dim(w)[2] == 3) {
    w <- cbind(w, 1 - w[, 1] - w[, 2] - w[, 3])
  }
  if (add.gauge) {
    wgrid <- create.wgrid.4d(20)
    ms <- additivegauge.scaling.4d(gauge1 = gauge1, gauge2 = gauge2,
                                   weight = par[length(par)], par1 = par[par1ind],
                                   par2 = par[par2ind], wgrid = wgrid)
    rate <- apply(w, 1, additivegauge.rescale.4d, par1 = par[par1ind],
                  par2 = par[par2ind], ms = ms, gauge1 = gauge1, gauge2 = gauge2,
                  weight = par[length(par)])
  }
  else {
    rate <- apply(w, 1, gfun, par = par[2:length(par)])
  }
  return(pgamma(k * r0w, shape = par[1], rate = rate, lower.tail = F)/pgamma(r0w,
                                                                             shape = par[1], rate = rate, lower.tail = F))
}

#==============================================================================

create.wgrid.4d = function (n)
{
  wseq <- seq(0, 1, len = n)
  wgrid <- expand.grid(wseq, wseq, wseq)
  wgrid <- cbind(wgrid, 1 - wgrid[, 1] - wgrid[, 2] - wgrid[, 3])
  neg.ind <- wgrid[, 4] < 0
  wgrid <- wgrid[!neg.ind, ]
  return(wgrid)
}

additivegauge.scaling.4d = function (gauge1, gauge2, par1, par2, weight, wgrid)
{
  dummy <- function(xyz) {
    gauge1(xyz, par = par1) + weight * gauge2(xyz, par = par2)
  }
  den <- apply(wgrid, 1, dummy)
  ms1 <- max(wgrid[, 1]/den)
  ms2 <- max(wgrid[, 2]/den)
  ms3 <- max(wgrid[, 3]/den)
  ms4 <- max(wgrid[, 4]/den)
  return(c(ms1, ms2, ms3, ms4))
}

additivegauge.rescale.4d = function (input.vec, gauge1, gauge2, par1, par2, weight, ms)
{
  input.vec <- c(ms[1] * input.vec[1], ms[2] * input.vec[2], ms[3] * input.vec[3], ms[4] * input.vec[4])
  return(gauge1(input.vec, par = par1) + weight * gauge2(input.vec, par = par2))
}


#########################################################

# JW added for consistency

#' Calculate P(R'>k|R'>1) (equivalently P(R>kr_0(W)|R>r_0(w)))

#' @param k value of k>1
#' @param w Values of W that correspond to the threshold exceedances of R
#' @param r0w The threshold function r_0(w) for each value of W supplied
#' @param gfun fitted gauge function
#' @param par fitted parameter values (first value should be the gamma shape parameter, subsequent values are the gauge function parameters)
#' @param add.gauge logical. If TRUE, then we additively mix (with numerical rescaling) the gauge functions specified by gauge1 and gauge2
#' @param gauge1 first gauge function for additive mixing. Use if add.gauge=TRUE
#' @param gauge2 second gauge function for additive mixing. Use if add.gauge=TRUE
#' @param par1ind vector of indices specifying the position of the parameters for gauge1 in par
#' @param par2ind vector of indices specifying the position of the parameters for gauge2 in par
#'
#' @return Estimate of P(R'>k|R'>1)
#'
#' @author Jenny Wadsworth
#' @export


Rexc.prob.k.4d<-function(k, r0w, w, gfun, par, add.gauge=FALSE,gauge1,gauge2,par1ind,par2ind)
{
  iw<-iweights.4d(k, r0w, w, gfun, par, add.gauge,gauge1,gauge2,par1ind,par2ind)
  return(mean(iw))
}


#' Use rolling windows to calculate local quantiles of R|W

#' @param r Values of R
#' @param w Values of W: should either be a two-column matrix containing (W1,W2) or a three-column matrix containing (W1,W2,W3)
#' @param method Character string: "empirical" for empirical rolling-windows quantiles. This is the only implemented method at present.
#' @param tau value in (0,1) - find the tau-quantile
#' @param J number of windows for rolling window estimate
#' @param wg Gap between the values of w to calculate the sizes of the rolling windows
#'
#' @return List with elements "wpts", an J^3 x 3 matrix containing the midpoints of the w1,w2,w3 values of each rolling window, "r.tau.wpts", a J x J x J array giving an estimate of the tau-quantile for R|W for each window, and "r0w", a threshold value of R|W for each value of w in the data
#'
#' @author Jenny Wadsworth
#' @export

QR.4d<-function(r,w,method="empirical",tau=0.95,J=20,wg=0.1)
{
  eqr<-empiricalQR.4d(r=r,w=w,tau=tau,J=J,wg=wg)
  wpts<-eqr$w.midpoint
  r.tau.wpts<-eqr$r.tau.w
  r0w<-eqr$r0w

  return(list(wpts=wpts, r.tau.wpts=r.tau.wpts, r0w=r0w))
}
